@@ -28,6 +28,8 @@
 
 #ifdef HAVE_GNUTLS
 #include <gnutls/gnutls.h>
+
+/* See comment in configure.ac */
 #ifdef HAVE_GNUTLS_SOCKET_H
 #include <gnutls/socket.h>
 #endif
@@ -161,6 +163,51 @@ nbd_unlocked_get_tls_username (struct nbd_handle *h)
   }
 }
 
+int
+nbd_unlocked_set_tls_hostname (struct nbd_handle *h, const char *hostname)
+{
+  char *new_hostname;
+
+  new_hostname = strdup (hostname);
+  if (!new_hostname) {
+    set_error (errno, "strdup");
+    return -1;
+  }
+  free (h->tls_hostname);
+  h->tls_hostname = new_hostname;
+  return 0;
+}
+
+static const char *
+get_tls_hostname (struct nbd_handle *h)
+{
+  if (h->tls_hostname)
+    return h->tls_hostname;
+  else if (h->hostname)
+    return h->hostname;
+  else
+    return NULL;
+}
+
+char *
+nbd_unlocked_get_tls_hostname (struct nbd_handle *h)
+{
+  const char *hostname = get_tls_hostname (h);
+  char *ret;
+
+  if (hostname)
+    ret = strdup (hostname);
+  else
+    /* Otherwise, return "" (not an error). */
+    ret = strdup ("");
+
+  if (ret == NULL) {
+    set_error (errno, "strdup");
+    return NULL;
+  }
+  return ret;
+}
+
 int
 nbd_unlocked_set_tls_psk_file (struct nbd_handle *h, const char *filename)
 {
@@ -176,6 +223,34 @@ nbd_unlocked_set_tls_psk_file (struct nbd_handle *h, const char *filename)
   return 0;
 }
 
+int
+nbd_unlocked_set_tls_priority (struct nbd_handle *h, const char *priority)
+{
+  char *new_priority;
+
+  new_priority = strdup (priority);
+  if (!new_priority) {
+    set_error (errno, "strdup");
+    return -1;
+  }
+  free (h->tls_priority);
+  h->tls_priority = new_priority;
+  return 0;
+}
+
+char *
+nbd_unlocked_get_tls_priority (struct nbd_handle *h)
+{
+  char *ret;
+
+  ret = strdup (h->tls_priority);
+  if (ret == NULL) {
+    set_error (errno, "strdup");
+    return NULL;
+  }
+  return ret;
+}
+
 #ifdef HAVE_GNUTLS
 
 static ssize_t
@@ -307,7 +382,7 @@ lookup_key (const char *pskfile, const char *username,
     if (r > ulen+1 &&
         strncmp (line, username, ulen) == 0 &&
         line[ulen] == ':') {
-      key->data = (unsigned char *) strdup (&line[ulen+1]);
+      key->data = (unsigned char *)strdup (&line[ulen+1]);
       if (key->data == NULL) {
         set_error (errno, "strdup");
         goto error;
@@ -338,21 +413,41 @@ lookup_key (const char *pskfile, const char *username,
   return -1;
 }
 
+static int
+set_priority (struct nbd_handle *h, gnutls_session_t session,
+              const char *priority)
+{
+  int err;
+
+  debug (h, "priority string: %s", priority);
+
+  err = gnutls_priority_set_direct (session, priority, NULL);
+  if (err < 0) {
+    set_error (0, "gnutls_priority_set_direct: %s: %s",
+               priority, gnutls_strerror (err));
+    return -1;
+  }
+  return 0;
+}
+
+
 static gnutls_psk_client_credentials_t
 set_up_psk_credentials (struct nbd_handle *h, gnutls_session_t session)
 {
   int err;
-  const char prio[] = TLS_PRIORITY ":" "+ECDHE-PSK:+DHE-PSK:+PSK";
+  char *prio = NULL;
   gnutls_datum_t key = { .data = NULL };
   char *username = NULL;
   gnutls_psk_client_credentials_t ret = NULL;
 
-  err = gnutls_priority_set_direct (session, prio, NULL);
-  if (err < 0) {
-    set_error (0, "gnutls_priority_set_direct: %s", gnutls_strerror (err));
+  if (asprintf (&prio, "%s:+ECDHE-PSK:+DHE-PSK:+PSK", h->tls_priority) == -1) {
+    set_error (errno, "strdup");
     goto error;
   }
 
+  if (set_priority (h, session, prio) == -1)
+    goto error;
+
   username = nbd_unlocked_get_tls_username (h);
   if (username == NULL)
     goto error;
@@ -380,11 +475,13 @@ set_up_psk_credentials (struct nbd_handle *h, gnutls_session_t session)
     goto error;
   }
 
+  free (prio);
   free (username);
   free (key.data);
   return ret;
 
  error:
+  free (prio);
   free (username);
   free (key.data);
   if (ret)
@@ -395,6 +492,7 @@ set_up_psk_credentials (struct nbd_handle *h, gnutls_session_t session)
 static int
 load_certificates (const char *path, gnutls_certificate_credentials_t *ret)
 {
+  size_t i;
   int err;
   char *cacert = NULL;
   char *clientcert = NULL;
@@ -439,6 +537,16 @@ load_certificates (const char *path, gnutls_certificate_credentials_t *ret)
     goto error;
   }
 
+  if (access (cacrl, R_OK) == 0) {
+    err = gnutls_certificate_set_x509_crl_file (*ret, cacrl,
+                                                GNUTLS_X509_FMT_PEM);
+    if (err < 0) {
+      set_error (0, "gnutls_certificate_set_x509_crl_file: %s: %s",
+                 cacrl, gnutls_strerror (err));
+      goto error;
+    }
+  }
+
   /* Optional for client certification authentication. */
   if (access (clientcert, R_OK) == 0 && access (clientkey, R_OK) == 0) {
     err = gnutls_certificate_set_x509_key_file (*ret, clientcert, clientkey,
@@ -450,12 +558,33 @@ load_certificates (const char *path, gnutls_certificate_credentials_t *ret)
     }
   }
 
-  if (access (cacrl, R_OK) == 0) {
-    err = gnutls_certificate_set_x509_crl_file (*ret, cacrl,
+  /* Load additional, optional client certificates and keys called
+   * 'client-cert-N.pem' and 'client-key-N.pem' for N = 0, 1, ...
+   * This is used by hybrid PQC.  See also:
+   * https://gitlab.com/berrange/gnutls-qemu-pqc-test-scenarios/
+   */
+  for (i = 0; ; ++i) {
+    free (clientcert);
+    clientcert = NULL;
+    if (asprintf (&clientcert, "%s/client-cert-%zu.pem", path, i) == -1) {
+      set_error (errno, "asprintf");
+      goto error;
+    }
+    free (clientkey);
+    clientkey = NULL;
+    if (asprintf (&clientkey, "%s/client-key-%zu.pem", path, i) == -1) {
+      set_error (errno, "asprintf");
+      goto error;
+    }
+
+    if (access (clientcert, R_OK) == -1 || access (clientkey, R_OK) == -1)
+      break;
+
+    err = gnutls_certificate_set_x509_key_file (*ret, clientcert, clientkey,
                                                 GNUTLS_X509_FMT_PEM);
     if (err < 0) {
-      set_error (0, "gnutls_certificate_set_x509_crl_file: %s: %s",
-                 cacrl, gnutls_strerror (err));
+      set_error (0, "gnutls_certificate_set_x509_key_file: %s, %s: %s",
+                 clientcert, clientkey, gnutls_strerror (err));
       goto error;
     }
   }
@@ -486,11 +615,8 @@ set_up_certificate_credentials (struct nbd_handle *h,
   const char *home = getenv ("HOME");
   char *path = NULL;
 
-  err = gnutls_priority_set_direct (session, TLS_PRIORITY, NULL);
-  if (err < 0) {
-    set_error (0, "gnutls_priority_set_direct: %s", gnutls_strerror (err));
+  if (set_priority (h, session, h->tls_priority) == -1)
     goto error;
-  }
 
   /* Try to load the certificates from the directory. */
   if (h->tls_certificates) {
@@ -532,13 +658,6 @@ set_up_certificate_credentials (struct nbd_handle *h,
   return NULL;
 
  found_certificates:
-#ifdef HAVE_GNUTLS_SESSION_SET_VERIFY_CERT
-  if (h->hostname && h->tls_verify_peer)
-    gnutls_session_set_verify_cert (session, h->hostname, 0);
-#else
-  debug (h, "ignoring nbd_set_tls_verify_peer, this requires GnuTLS >= 3.4.6");
-#endif
-
   err = gnutls_credentials_set (session, GNUTLS_CRD_CERTIFICATE, ret);
   if (err < 0) {
     set_error (0, "gnutls_credentials_set: %s", gnutls_strerror (err));
@@ -561,11 +680,8 @@ set_up_system_CA (struct nbd_handle *h, gnutls_session_t session)
   int err;
   gnutls_certificate_credentials_t ret = NULL;
 
-  err = gnutls_priority_set_direct (session, TLS_PRIORITY, NULL);
-  if (err < 0) {
-    set_error (0, "gnutls_priority_set_direct: %s", gnutls_strerror (err));
+  if (set_priority (h, session, h->tls_priority) == -1)
     return NULL;
-  }
 
   err = gnutls_certificate_allocate_credentials (&ret);
   if (err < 0) {
@@ -653,6 +769,14 @@ nbd_internal_crypto_create_session (struct nbd_handle *h,
       gnutls_deinit (session);
       return NULL;
     }
+
+    if (h->tls_verify_peer) {
+      /* Verify the server certificate using h->tls_hostname, falling
+       * back to h->hostname.
+       */
+      const char *hostname = get_tls_hostname (h);
+      gnutls_session_set_verify_cert (session, hostname, 0);
+    }
   }
 
   /* Wrap the underlying socket with GnuTLS. */
@@ -712,7 +836,7 @@ nbd_internal_crypto_handshake (struct nbd_handle *h)
   in = gnutls_handshake_get_last_in (session);
   out = gnutls_handshake_get_last_out (session);
   set_error (0, "gnutls_handshake: %s (%d/%d)",
-             gnutls_strerror (err), (int) in, (int) out);
+             gnutls_strerror (err), (int)in, (int)out);
   return -1;
 }
 
@@ -742,7 +866,7 @@ nbd_internal_crypto_debug_tls_enabled (struct nbd_handle *h)
     case GNUTLS_KTLS_SEND: ktls_status = "enabled send only"; break;
     case GNUTLS_KTLS_DUPLEX: ktls_status = "enabled"; break;
     default:
-      if ((int) ktls_enabled == 0)
+      if ((int)ktls_enabled == 0)
         ktls_status = "disabled";
       else
         ktls_status = "unknown";
