@@ -38,6 +38,8 @@
 #endif
 
 #include "array-size.h"
+#include "const-string-vector.h"
+#include "parse-bool.h"
 #include "vector.h"
 
 #include "internal.h"
@@ -160,29 +162,176 @@ error:
   return -1;
 }
 
-/* Similar to nbdkit_parse_bool */
 static int
-parse_bool (const char *param, const char *value)
+parse_bool_param (const char *param, const char *value)
 {
-  if (!strcmp (value, "1") ||
-      !strcasecmp (value, "true") ||
-      !strcasecmp (value, "t") ||
-      !strcasecmp (value, "yes") ||
-      !strcasecmp (value, "y") ||
-      !strcasecmp (value, "on"))
-    return 1;
+  int r = parse_bool (value);
+  if (r == -1) {
+    set_error (EINVAL, "could not parse %s parameter, "
+               "expecting %s=true|false",
+               param, param);
+    return -1;
+  }
+  return r;
+}
 
-  if (!strcmp (value, "0") ||
-      !strcasecmp (value, "false") ||
-      !strcasecmp (value, "f") ||
-      !strcasecmp (value, "no") ||
-      !strcasecmp (value, "n") ||
-      !strcasecmp (value, "off"))
-    return 0;
+static int
+uri_connect_tcp (struct nbd_handle *h, xmlURIPtr uri, const char *nbd_port)
+{
+  char *server;
+  size_t server_len;
+  char port_str[32];
+  const char *tcp_port;
 
-  set_error (EINVAL, "could not parse %s parameter, expecting %s=true|false",
-             param, param);
-  return -1;
+  /* Prefer nbd-port if used, else use the URI port. */
+  if (nbd_port)
+    tcp_port = nbd_port;
+  else {
+    snprintf (port_str, sizeof port_str,
+              "%d", uri->port > 0 ? uri->port : 10809);
+    tcp_port = port_str;
+  }
+
+  /* If the uri->server field is NULL, substitute "localhost".  This
+   * would be unusual and probably doesn't happen in reality. XXX
+   */
+  server = uri->server ? : "localhost";
+  server_len = strlen (server);
+
+  /* For a literal IPv6 address, the uri->server field will contain
+   * "[addr]" and we must remove the brackets before passing it to
+   * getaddrinfo in the state machine.  We treat uri->server as
+   * writable here.
+   */
+  if (server_len >= 2 && server[0] == '[' && server[server_len-1] == ']') {
+    server_len -= 2;
+    server++;
+    server[server_len] = '\0';
+  }
+
+  if (nbd_unlocked_aio_connect_tcp (h, server, tcp_port) == -1)
+    return -1;
+
+  return 0;
+}
+
+static int
+uri_connect_vsock (struct nbd_handle *h, const xmlURIPtr uri,
+                   const char *nbd_port)
+{
+  uint32_t cid, svm_port;
+
+  /* Server, if present, must be the numeric CID.  Else we
+   * assume the host (2).
+   */
+  if (uri->server && strcmp (uri->server, "") != 0) {
+    /* This doesn't deal with overflow, but that seems unlikely to
+     * matter because you'll end up with a CID one way or another.
+     */
+    if (sscanf (uri->server, "%" SCNu32, &cid) != 1) {
+      set_error (EINVAL, "cannot parse vsock CID from NBD URI: %s",
+                 uri->server);
+      return -1;
+    }
+  }
+  else
+    cid = 2;
+
+  if (nbd_port) {
+    if (sscanf (nbd_port, "%" SCNu32, &svm_port) != 1) {
+      set_error (EINVAL, "cannot parse vsock port from nbd-port: %s",
+                 nbd_port);
+      return -1;
+    }
+  }
+  else {
+    /* For unknown reasons libxml2 sets uri->port = -1 if the
+     * authority field is not present at all.  So we must check that
+     * uri->port > 0.
+     */
+    svm_port = uri->port > 0 ? (uint32_t)uri->port : 10809;
+  }
+
+  if (nbd_unlocked_aio_connect_vsock (h, cid, svm_port) == -1)
+    return -1;
+
+  return 0;
+}
+
+static int
+uri_connect_ssh (struct nbd_handle *h, const xmlURIPtr uri,
+                 const char *nbd_port, const char *unixsocket,
+                 bool ssh_compress)
+{
+  const_string_vector ssh_command = empty_vector;
+  char *ssh_user = NULL;
+  char port_str[32];
+  int ret = -1;
+
+  if (!uri->server || strcmp (uri->server, "") == 0) {
+    set_error (EINVAL, "SSH transport requires a server name");
+    goto cleanup;
+  }
+
+  /* Reserve enough space for the longest SSH command, including
+   * final NULL pointer.
+   */
+  if (const_string_vector_reserve (&ssh_command, 16) == -1) {
+    set_error (errno, "realloc");
+    goto cleanup;
+  }
+
+  const_string_vector_append (&ssh_command, "ssh");
+  if (ssh_compress)
+    const_string_vector_append (&ssh_command, "-C");
+
+  /* SSH port always comes from the authority part of the URI. */
+  snprintf (port_str, sizeof port_str,
+            "%d", uri->port > 0 ? uri->port : 22);
+  const_string_vector_append (&ssh_command, "-p");
+  const_string_vector_append (&ssh_command, port_str);
+
+  /* SSH user always comes from the authority part of the URI. */
+  if (uri->user) {
+    /* Don't allow any special tokens in the username which ssh might
+     * interpret.
+     */
+    if (strchr (uri->user, '%')) {
+      set_error (EINVAL, "invalid SSH username, cannot contain '%%'");
+      goto cleanup;
+    }
+    if (asprintf (&ssh_user, "User=%s", uri->user) == -1) {
+      set_error (errno, "asprintf");
+      goto cleanup;
+    }
+    const_string_vector_append (&ssh_command, "-o");
+    const_string_vector_append (&ssh_command, ssh_user);
+  }
+
+  const_string_vector_append (&ssh_command, "--");
+  const_string_vector_append (&ssh_command, uri->server);
+
+  /* All further argc beginning with - are options to nc. */
+  const_string_vector_append (&ssh_command, "nc");
+  if (unixsocket) {
+    const_string_vector_append (&ssh_command, "-U");
+    const_string_vector_append (&ssh_command, unixsocket);
+  }
+  else {
+    const_string_vector_append (&ssh_command, "localhost");
+    const_string_vector_append (&ssh_command, nbd_port ? : "10809");
+  }
+  const_string_vector_append (&ssh_command, NULL);
+
+  if (nbd_unlocked_aio_connect_command (h, (char **) ssh_command.ptr) == -1)
+    goto cleanup;
+
+  ret = 0;
+
+ cleanup:
+  const_string_vector_reset (&ssh_command);
+  free (ssh_user);
+  return ret;
 }
 
 int
@@ -195,7 +344,10 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
   uri_query_list queries = empty_vector;
   int i, r;
   int ret = -1;
+  const char *nbd_port = NULL;
+  const char *tls_username = NULL;
   const char *unixsocket = NULL;
+  bool ssh_compress = false;
 
   uri = xmlParseURI (raw_uri);
   if (!uri) {
@@ -332,9 +484,18 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
   if (tls && nbd_unlocked_set_tls (h, LIBNBD_TLS_REQUIRE) == -1)
     goto cleanup;
 
-  /* Look for some tls-* parameters. */
+  /* Look for some other query parameters. */
   for (i = 0; i < queries.len; i++) {
-    if (strcasecmp (queries.ptr[i].name, "tls-certificates") == 0) {
+    if (strcasecmp (queries.ptr[i].name, "nbd-port") == 0) {
+      nbd_port = queries.ptr[i].value;
+    }
+    else if (strcasecmp (queries.ptr[i].name, "compress") == 0) {
+      int v = parse_bool_param ("compress", queries.ptr[i].value);
+      if (v == -1)
+        goto cleanup;
+      ssh_compress = v;
+    }
+    else if (strcasecmp (queries.ptr[i].name, "tls-certificates") == 0) {
       if (! h->uri_allow_local_file) {
         set_error (EPERM,
                    "local file access (tls-certificates) is not allowed, "
@@ -358,8 +519,21 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
       if (nbd_unlocked_set_tls_hostname (h, queries.ptr[i].value) == -1)
         goto cleanup;
     }
+    else if (strcasecmp (queries.ptr[i].name, "tls-priority") == 0) {
+      if (! h->uri_allow_tls_priority) {
+        set_error (EPERM,
+                   "overriding TLS priority (tls-priority) is not allowed, "
+                   "call nbd_set_uri_allow_tls_priority to enable this");
+        goto cleanup;
+      }
+      if (nbd_unlocked_set_tls_priority (h, queries.ptr[i].value) == -1)
+        goto cleanup;
+    }
+    else if (strcasecmp (queries.ptr[i].name, "tls-username") == 0) {
+      tls_username = queries.ptr[i].value; /* set below */
+    }
     else if (strcasecmp (queries.ptr[i].name, "tls-verify-peer") == 0) {
-      int v = parse_bool ("tls-verify-peer", queries.ptr[i].value);
+      int v = parse_bool_param ("tls-verify-peer", queries.ptr[i].value);
       if (v == -1)
         goto cleanup;
       if (nbd_unlocked_set_tls_verify_peer (h, v) == -1)
@@ -367,8 +541,10 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
     }
   }
 
-  /* Username. */
-  if (uri->user && nbd_unlocked_set_tls_username (h, uri->user) == -1)
+  /* Set the TLS username.  Always prefer tls-username parameter. */
+  if (!tls_username)
+    tls_username = uri->user;
+  if (tls_username && nbd_unlocked_set_tls_username (h, tls_username) == -1)
     goto cleanup;
 
   /* Export name. */
@@ -382,103 +558,26 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
   if (r == -1)
     goto cleanup;
 
+  /* Do the connection. */
   switch (transport) {
-  case tcp: {                   /* TCP */
-    char port_str[32];
-    char *server;
-    size_t server_len;
-
-    snprintf (port_str, sizeof port_str,
-              "%d", uri->port > 0 ? uri->port : 10809);
-
-    /* If the uri->server field is NULL, substitute "localhost".  This
-     * would be unusual and probably doesn't happen in reality. XXX
-     */
-    server = uri->server ? : "localhost";
-    server_len = strlen (server);
-
-    /* For a literal IPv6 address, the uri->server field will contain
-     * "[addr]" and we must remove the brackets before passing it to
-     * getaddrinfo in the state machine.
-     */
-    if (server_len >= 2 && server[0] == '[' && server[server_len-1] == ']') {
-      server_len -= 2;
-      server++;
-      server[server_len] = '\0';
-    }
-
-    if (nbd_unlocked_aio_connect_tcp (h, server, port_str) == -1)
+  case tcp:                     /* TCP */
+    if (uri_connect_tcp (h, uri, nbd_port) == -1)
       goto cleanup;
-
     break;
-  }
 
   case unix_sock:               /* Unix domain socket */
     if (nbd_unlocked_aio_connect_unix (h, unixsocket) == -1)
       goto cleanup;
-
     break;
 
-  case vsock: {                 /* AF_VSOCK */
-    uint32_t cid, svm_port;
-
-    /* Server, if present, must be the numeric CID.  Else we
-     * assume the host (2).
-     */
-    if (uri->server && strcmp (uri->server, "") != 0) {
-      /* This doesn't deal with overflow, but that seems unlikely to
-       * matter because you'll end up with a CID one way or another.
-       */
-      if (sscanf (uri->server, "%" SCNu32, &cid) != 1) {
-        set_error (EINVAL, "cannot parse vsock CID from NBD URI: %s",
-                   uri->server);
-        goto cleanup;
-      }
-    }
-    else
-      cid = 2;
-
-    /* For unknown reasons libxml2 sets uri->port = -1 if the
-     * authority field is not present at all.  So we must check that
-     * uri->port > 0.  This prevents us from using certain very large
-     * port numbers, but that's not an issue that matters in practice.
-     */
-    svm_port = uri->port > 0 ? (uint32_t)uri->port : 10809;
-    if (nbd_unlocked_aio_connect_vsock (h, cid, svm_port) == -1)
+  case vsock:                   /* AF_VSOCK */
+    if (uri_connect_vsock (h, uri, nbd_port) == -1)
       goto cleanup;
-
     break;
-  }
 
-  case ssh: {                   /* SSH */
-    char port_str[32];
-    const char *ssh_command[] = {
-      "ssh", "-p", port_str, uri->server,
-      "nc",
-      NULL,                     /* [5] "-U" or "localhost" */
-      NULL,                     /* [6] socket or "10809" */
-      NULL,
-    };
-
-    if (!uri->server || strcmp (uri->server, "") == 0) {
-      set_error (EINVAL, "SSH transport requires a server name");
+  case ssh:                     /* SSH */
+    if (uri_connect_ssh (h, uri, nbd_port, unixsocket, ssh_compress) == -1)
       goto cleanup;
-    }
-    snprintf (port_str, sizeof port_str,
-              "%d", uri->port > 0 ? uri->port : 22);
-
-    if (unixsocket) {
-      ssh_command[5] = "-U";
-      ssh_command[6] = unixsocket;
-    }
-    else {
-      ssh_command[5] = "localhost";
-      ssh_command[6] = "10809"; /* XXX provide a way to configure this */
-    }
-
-    if (nbd_unlocked_aio_connect_command (h, (char **) ssh_command) == -1)
-      goto cleanup;
-  }
   }
 
   ret = 0;
@@ -642,6 +741,13 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
                              "tls-hostname", h->tls_hostname) == -1)
       goto out;
   }
+  if (h->uri_allow_tls_priority && h->tls_priority &&
+      /* tls-priority is only added to URIs if it has a non-default value. */
+      strcmp (h->tls_priority, TLS_PRIORITY) != 0) {
+    if (append_query_params (&query_params,
+                             "tls-priority", h->tls_priority) == -1)
+      goto out;
+  }
   if (! h->tls_verify_peer) {
     if (append_query_params (&query_params,
                              "tls-verify-peer", "false") == -1)
